/*
 * Copyright 2012 Stefan Haun, Andreas Nürnberger
 * 
 *      Data and Knowledge Engineering Group, 
 * 		Faculty of Computer Science,
 *		Otto-von-Guericke University,
 *		Magdeburg, Germany
 *
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.ovgu.dke.mocca.daemon;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;

import org.apache.commons.daemon.Daemon;
import org.apache.commons.daemon.DaemonContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.ovgu.dke.mocca.api.MoccaException;
import de.ovgu.dke.mocca.api.MoccaRuntime;
import de.ovgu.dke.mocca.api.command.CommandHandler;
import de.ovgu.dke.mocca.util.MoccaHelper;

/**
 * MOCCA implementation of the JSVC daemon.
 * 
 * @author Stefan Haun (stefan.haun@ovgu.de)
 */
public class MoccaDaemon implements Daemon {
    static Log log = LogFactory.getLog(MoccaDaemon.class);

    private final Collection<MoccaAgent> agents;
    private DaemonContext ctx;

    protected enum State {
        Created, Initialized, Running, Destroyed, Failed
    }

    private State state;

    public MoccaDaemon() {
        this.agents = new ArrayList<MoccaAgent>();
        this.state = State.Created;
    }

    public State getState() {
        synchronized (state) {
            return state;
        }
    }

    protected void setState(final State state) {
        synchronized (this.state) {
            final State _state = this.state;
            this.state = state;
            log.info("Mocca Daemon switched state from " + _state + " to "
                    + this.state + ".");
        }
    }

    @Override
    public void init(DaemonContext ctx) throws Exception {
        if (ctx == null)
            throw new NullPointerException("Daemon Context may not be null!");

        synchronized (this.state) {
            this.ctx = ctx;

            // TODO get agents from configuration

            setState(State.Initialized);
        }
    }

    /**
     * Add a Mocca Agent to the daemon. This must be done during the
     * initialization state, i.e. the service must either be freshly initialized
     * or stopped.
     * 
     * @param agent
     *            the agent instance to be added.
     * @throws MoccaException
     *             in case of an invalid state.
     */
    public void addAgent(final MoccaAgent agent) throws MoccaException {
        synchronized (state) {
            if (!state.equals(State.Initialized))
                throw new MoccaException(
                        "Agents may only be added if Mocca Daemon is in state Initialized!");

            this.agents.add(agent);
        }

    }

    /**
     * Load an agent from the provided class path and add it to the daemon. The
     * rules for instantiating and adding agents apply!
     * 
     * @param classPath
     *            Fully qualified class path of the Mocca Agent.
     * @throws MoccaException
     *             if the class cannot be found or instantiated.
     */
    public void loadAgent(final String classPath) throws MoccaException {
        try {
            final Class<?> daemonClass = Class.forName(classPath);
            loadAgent(daemonClass);
        } catch (ClassNotFoundException e) {
            throw new MoccaException("Daemon class not found: "
                    + e.getMessage(), e);
        }
    }

    /**
     * Load an agent from the provided class and add it to the daemon. The agent
     * needs to have a public constructor with no parameters. The rules for
     * adding agents apply!
     * 
     * @param daemonClass
     *            The Mocca Agent class to be loaded.
     * @throws MoccaException
     *             if the class cannot be instantiated.
     */
    public void loadAgent(final Class<?> daemonClass) throws MoccaException {
        Constructor<?> con;
        try {
            con = daemonClass.getConstructor();
            final MoccaAgent agent = (MoccaAgent) con.newInstance();

            this.addAgent(agent);
        } catch (Exception e) {
            throw new MoccaException("Error on loading agent from class: "
                    + e.getMessage(), e);
        }

    }

    @Override
    public void destroy() {
        synchronized (state) {
            if (state.equals(State.Running)) {
                ctx.getController().fail(
                        "Mocca Daemon still running, will not Destroy!");
                return;
            }

            for (final MoccaAgent agent : agents) {
                log.info("Destroying agent " + agent.getName());
                agent.destroy();
            }

            setState(State.Destroyed);
        }
    }

    @Override
    public void start() throws Exception {
        synchronized (this.state) {
            if (getState().equals(State.Running)) {
                log.warn("Tried to start Mocca Daemon which is already Running! Ignored.");
                return;
            }

            if (!getState().equals(State.Initialized))
                throw new MoccaException(
                        "Mocca Daemon may only be started if in state Initialized!");

            // init mocca runtime
            log.info("Starting Mocca Runtime");
            final MoccaRuntime runtime = MoccaHelper.getDefaultRuntime();
            runtime.init(MoccaRuntime.NULL_ENV);

            // init agents
            for (final MoccaAgent agent : agents) {
                log.info("Registering Mocca Agent " + agent.getName());
                try {
                    initAgent(agent, runtime);
                } catch (MoccaException e) {
                    log.error(
                            "Error initializing mocca agent: " + e.getMessage(),
                            e);
                }

            }

            setState(State.Running);
        }
    }

    private void initAgent(final MoccaAgent agent, final MoccaRuntime runtime)
            throws MoccaException {
        agent.init(runtime);

        final Collection<CommandHandler> handlers = agent.getCommandHandlers();
        if (handlers != null)
            for (final CommandHandler handler : handlers)
                MoccaHelper.getDefaultRuntime().getCommandHandlerRegistry()
                        .registerCommandHandler(handler);
    }

    @Override
    public void stop() throws Exception {
        synchronized (this.state) {
            if (!getState().equals(State.Running)) {
                ctx.getController()
                        .fail("Tried to stop Mocca Daemon which is not Running! Ignored.");
                return;
            }

            // dispose agents
            for (final MoccaAgent agent : agents)
                agent.destroy();

            // dispose mocca runtime
            log.info("Disposing Mocca Runtime");
            MoccaHelper.getDefaultRuntime().dispose();

            setState(State.Initialized);
        }
    }

}
